#ifndef DLIST_H
#define DLIST_H

/* Doubly linked list.

   This implementation of a doubly linked list does not require
   use of dynamically allocated memory.  Instead, each structure
   that is a potential list element must embed a struct DNode
   member.  All of the list functions operate on these struct
   DNode's.  The DList_Entry macro allows conversion from a
   struct DNode back to a structure object that contains it.

   For example, suppose there is a need for a list of 'struct Foo'.
   'struct Foo' should contain a `struct DNode' member, like so:

      struct Foo {
         struct DNode node;
         int bar;
         ...other members...
      };

   Then a list of 'struct Foo' can be be declared and initialized
   like so:

      struct DList Foo_List;

      DList_Init (&Foo_List);

   Iteration is a typical situation where it is necessary to
   convert from a struct DNode back to its enclosing
   structure.  Here's an example using Foo_List:

      struct DNode *e;

      for (e = DList_Begin (&Foo_List); e != DList_End (&Foo_List); e = DList_Next (e)) {

         struct Foo *f = DList_Entry (e, struct Foo, node);
         ...do something with f...
      }

   The interface for this list is inspired by the list<> template
   in the C++ STL.  If you're familiar with list<>, you should
   find this easy to use.  However, it should be emphasized that
   these lists do *no* type checking and can't do much other
   correctness checking.  If you screw up, it will bite you.

   Glossary of list terms:

     - "front": The first element in a list.  Undefined in an
       empty list.  Returned by DList_Front().

     - "back": The last element in a list.  Undefined in an empty
       list.  Returned by DList_Back().

     - "tail": The element figuratively just after the last
       element of a list.  Well defined even in an empty list.
       Returned by DList_End().  Used as the end sentinel for an
       iteration from front to back.

     - "beginning": In a non-empty list, the front.  In an empty
       list, the tail.  Returned by DList_Begin().  Used as the
       starting point for an iteration from front to back.

     - "head": The element figuratively just before the first
       element of a list.  Well defined even in an empty list.
       Returned by DList_Rend().  Used as the end sentinel for an
       iteration from back to front.

     - "reverse beginning": In a non-empty list, the back.  In an
       empty list, the head.  Returned by DList_Rbegin().  Used as
       the starting point for an iteration from back to front.

     - "interior element": An element that is not the head or
       tail, that is, a real list element.  An empty list does
       not have any interior elements.
*/

#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

/* List node. */
struct _DNode {

   struct _DNode *prev;     /* Previous list element. */
   struct _DNode *next;     /* Next list element.    */
};

/* List. */
struct _DList {

   struct _DNode head;      /* List head. */
   struct _DNode tail;      /* List tail. */
};

typedef struct _DNode DNode;
typedef struct _DList DList;

/* Converts pointer to list element LIST_ELEM into a pointer to
   the structure that LIST_ELEM is embedded inside.  Supply the
   name of the outer structure STRUCT and the member name MEMBER
   of the list element.  See the big comment at the top of the
   file for an example. */
#define DList_Entry(LIST_ELEM, STRUCT, MEMBER)                              \
        ((STRUCT *) ((uint8_t *) (LIST_ELEM) - offsetof (STRUCT, MEMBER)))

void DList_Init(DList* pL);

void DList_PushFront(DList* pL, DNode* node);
void DList_PushBack(DList* list, DNode* elem);
void DList_Insert(DNode* before, DNode* elem);

DNode* DList_Remove(DNode* elem);
DNode* DList_PopFront(DList* list);
DNode* DList_PopBack(DList* list);

bool DList_Empty(DList* list);

DNode* DList_Begin(DList* pL);
DNode* DList_Next(DNode* elem);
DNode* DList_End(DList* list);
DNode* DList_RBegin(DList* list);
DNode* DList_Prev(DNode* elem);
DNode* DList_REnd(DList* list);
DNode* DList_Head(DList* list);
DNode* DList_Tail(DList* list);
DNode* DList_Front(DList* list);
DNode* DList_Back(DList* list);

#endif

